package edu.berkeley.qfat.geom;
import java.util.*;

/** a set of points, plus many useful methods operating over the set */
public class PointSet<V extends HasPoint> implements Iterable<V> {

    private RTree<V> rtree = new RTree<V>();

    private HashMap<Point,V> exact = new HashMap<Point,V>();

    public int size() { return exact.size(); }

    public Iterator<V> iterator() { return rtree.iterator(); }

    public void clear() {
        exact.clear();
        rtree = new RTree<V>();
    }

    public V get(Point p) {
        return exact.get(p);
    }

    public void add(V v) {
        V x = get(v.getPoint());
        if (x != null && x.equals(v)) return;
        if (x != null) throw new Error("duplicates!");
        Point p = v.getPoint();
        rtree.insert(v);
        exact.put(p, v);
    }

    public void remove(V v) {
        Point p = v.getPoint();
        rtree.remove(v);
        exact.remove(p);
    }

    public V nearest(Point p) { return nearest(p, null); }
    public V nearest(Point p, Visitor<V> vis) {
        if (exact.size()==0) return null;
        return rtree.nearest(p, vis);
    }

    public Vec diagonal() {
        float min_x = Float.MAX_VALUE;
        float min_y = Float.MAX_VALUE;
        float min_z = Float.MAX_VALUE;
        float max_x = Float.MIN_VALUE;
        float max_y = Float.MIN_VALUE;
        float max_z = Float.MIN_VALUE;
        for(V v : this) {
            Point p = v.getPoint();
            if (p.x < min_x) min_x = p.x;
            if (p.y < min_y) min_y = p.y;
            if (p.z < min_z) min_z = p.z;
            if (p.x > max_x) max_x = p.x;
            if (p.y > max_y) max_y = p.y;
            if (p.z > max_z) max_z = p.z;
        }
        return new Vec(max_x - min_x, max_y - min_y, max_z - min_z);
    }

    // FEATURE: compute incrementally?
    public Point centroid() {
        float min_x = Float.MAX_VALUE;
        float min_y = Float.MAX_VALUE;
        float min_z = Float.MAX_VALUE;
        float max_x = Float.MIN_VALUE;
        float max_y = Float.MIN_VALUE;
        float max_z = Float.MIN_VALUE;
        for(V v : this) {
            Point p = v.getPoint();
            if (p.x < min_x) min_x = p.x;
            if (p.y < min_y) min_y = p.y;
            if (p.z < min_z) min_z = p.z;
            if (p.x > max_x) max_x = p.x;
            if (p.y > max_y) max_y = p.y;
            if (p.z > max_z) max_z = p.z;
        }
        return new Point((float)(max_x + min_x)/2,
                     (float)(max_y + min_y)/2,
                     (float)(max_z + min_z)/2);
    }
}
